// clang-format off
/*
 * SPDX-FileCopyrightText: Copyright (c) 2023-present NVIDIA CORPORATION & AFFILIATES.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 */
// clang-format on
#pragma once

#include <device_lower/analysis/trivial_broadcast.h>
#include <id_model/id_model.h>
#include <ir/base_nodes.h>
#include <ir/interface_nodes.h>
#include <options.h>
#include <type.h>

// Just for PredicateInfo. Should be moved to its own header file
#include <index_compute.h>

#include <unordered_map>

namespace nvfuser {

struct IndexingInfo {
  std::vector<IterDomain*> loop_domains;
  // Indexing traversal path from loop domains
  ExprPath<ExprGroup> traversal_path;
  // Index mappings of ID groups along the traversal path
  std::unordered_map<ValGroup, Val*> index_map;
  // Mappings from ID groups to dependent loop groups
  std::unordered_map<ValGroup, ValGroups> loop_group_dependencies;
};

struct IndexingAllocationInfo {
  std::vector<IterDomain*> domains;
  std::vector<Val*> strides;
  std::vector<bool> contiguity;
};

// The basic algorithm of indexing is:
//
// 1. Find the loop domains
// 2. Find the allocation domains
// 3. Find the path from the loop domains to the allocation domains
// 4. Set the initial index vals for the loop domains
// 5. Propagate the initial indices of the loop domains to the allocation
// domains
//
// The indexing traversal is done on the AlmostExact graph augmented
// with the loop promotion map since both the loop and allocations
// domains may be promoted.
class TensorIndexer {
 public:
  // Using non-const references of IdModel because traversalGraph() returns a
  // non-const reference
  TensorIndexer(IdModel& id_model);

  bool isContigIndexingEnabled() const {
    return !isOptionDisabled(DisableOption::ContigIndexing);
  }

  // Get a linear index of a given tensor appearing in a given expr, either
  // as a consumer or a producer. The predicate indexing will have a
  // separate interface.
  //
  // The actual for-loops are required for handling circular buffering
  Val* getLinearIndex(
      TensorView* tv,
      const Expr* expr,
      const std::vector<ForLoop*>& loops) const;

  // Get the index of a loop domain.
  Val* getLoopIndex(IterDomain* loop_id, const std::vector<ForLoop*>& for_loops)
      const;

  // Get the index of the given ID groups
  std::vector<Val*> getIndexFor(
      const Expr* expr,
      bool as_consumer,
      const std::vector<IterDomain*>& index_ids,
      const std::vector<ForLoop*>& loops) const;

  // Get the contig indices of the given ID groups with their strides
  std::pair<std::vector<Val*>, std::vector<Val*>> getContigIndexFor(
      const Expr* expr,
      bool as_consumer,
      const IndexingAllocationInfo& alloc_info,
      const std::vector<ForLoop*>& loops) const;

  // The AlmostExact graph is used since size-1 splits and merges
  // should not affect actual index exprs.
  // Returns non-const reference because indexing may create new domains and
  // need to update the graph.

  static IdMappingMode traversalGraphType() {
    return IdMappingMode::ALMOSTEXACT;
  }

  ValGraph& traversalGraph() const {
    return id_model_.idGraph(traversalGraphType());
  }

  // Traverse exprs and set allocation info for each tensor
  void setupAllocationDomains(const std::vector<Expr*>& exprs);

  // Get the list of predicates of a given tensor appearing in a given
  // expr as a consumer. Each predicate corresponds to a domain of the
  // tensor, which is by default one of the logical domains but can be
  // an intermediate domain with contiguous indexing.
  //
  // An optional ForLoop parameter specifies a loop that is either
  // unswitched/unrolled or vectorized, both of which are handled by
  // UnswitchPredicate. For normal inline predicates, the parameter
  // should be nullptr.
  std::vector<PredicateInfo> getPredicates(
      TensorView* tv,
      const Expr* expr,
      const std::vector<ForLoop*>& for_loops,
      ForLoop* unswitched_loop = nullptr) const;

  // Get the indexing traversal path for indexing a given list of IDs
  // for a given expr
  ExprPath<ExprGroup> getIndexingPath(
      const Expr* expr,
      const std::vector<IterDomain*>& index_ids) const;

 private:
  // Build a map of loop groups to their index Vals. See the comment
  // on loop_index_map_.
  void buildLoopIndexMap();

  const IndexingAllocationInfo& getIndexingAllocationInfo(
      TensorView* tv) const {
    auto it = alloc_info_.find(tv);
    NVF_ERROR(
        it != alloc_info_.end(),
        "No allocation info found for ",
        tv->toString());
    return it->second;
  }

  // Returns the index map as well as its traversal path of given
  // index domains appearing in a given expr. Used by
  // getIndexFor.
  IndexingInfo computeIndex(
      const Expr* expr,
      const std::vector<IterDomain*>& index_ids,
      const std::vector<ForLoop*>& for_loops) const;

  // Propagate the loop indices of a given list of loop domains to the
  // traversal graph (i.e., the AlmostExact graph). Uses the loop
  // index map, which is built for the Loop graph.
  std::unordered_map<ValGroup, Val*> getInitialIndexMap(
      const std::vector<IterDomain*>& loop_domains,
      const std::vector<ForLoop*>& for_loops) const;

  // Get the loop domains of a given expr. Currently, they're always
  // the loop domains of a consumer tensor, but in the future this
  // function may return the loop domains of a producer for
  // producer-based indexing.
  std::vector<IterDomain*> getLoopDomains(const Expr* expr) const;

  // For a given indexng traversal path toward allocation_domains,
  // return the contiguous domains and their strides that can provide
  // equivalent indexing results.
  //
  // Currently, only backward traversal is supported.
  std::pair<std::vector<ValGroup>, std::vector<Val*>> getContigDomainsAndStrides(
      const IndexingAllocationInfo& alloc_info,
      const ExprPath<ExprGroup>& traversal_path) const;

  // Get a replace map for tensor indexing. Examples include replacing
  // an index of a vectorized loop with zero.
  //
  // This replacement map is used to replace a tensor index after an
  // index map is generated. Since replacment is only done for loop
  // domains, this could be done as part of getInitialIndexMap. One
  // reason that we might want to first generate an index and do some
  // replacements, rather than using final index vals to build the
  // index map, is that one index map could be used for multiple
  // indices. For normal tensor indexing, this may not matter, but for
  // predicate indexing, it needs to generate both start and stop
  // predicates, and one index map would be sufficient for both
  // indices by using different replacement maps.
  std::unordered_map<Val*, Val*> getIndexReplacementMap(
      const Expr* expr,
      bool as_consumer,
      const std::vector<IterDomain*>& loop_domains,
      const std::vector<ForLoop*>& for_loops,
      const std::unordered_map<ValGroup, Val*>& index_map) const;

 private:
  // Using non-const references of IdModel because traversalGraph() returns a
  // non-const reference
  IdModel& id_model_;

  // Mappings from loop groups to their indices. Serial loops will
  // be mapped a unique loop index Val. Parallel loops will be mapped
  // to NamedScalar such as "threadIdx.x". This map needs to be built
  // once and can be reused for different tensors.
  std::unordered_map<ValGroup, Val*> loop_index_map_;

  // Allocation info for each tensor. Must be filled before computing
  // the index of each tensor
  std::unordered_map<TensorView*, IndexingAllocationInfo> alloc_info_;
};

} // namespace nvfuser
